#ifndef GEOGRAPHY
#define GEOGRAPHY

double degrees_lat_to_meters(double dy, double y);
double degrees_lon_to_meters(double dx, double y);
double meters_to_degrees_lon(double dx_meters, double y);
double meters_to_degrees_lat(double dy_meters, double y);
double constrain_longitude_to_valid_domain(double lon);

#endif // GEOGRAPHY
