#ifndef PHYSICAL_CONSTANTS
#define PHYSICAL_CONSTANTS

#define ACC_GRAVITY -9.81  // m s^-2
#define DYNAMIC_VISCOSITY_SEAWATER 0.0012  // kg m^-1 s^-1, typical value (ITTC fresh water and seawater properties,
                                            // https://ittc.info/media/4048/75-02-01-03.pdf)

#endif // PHYSICAL_CONSTANTS
